@@ -9,12 +9,14 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <string.h>
 #endif
 
 #include <iostream>
 #include <gensio/gensio>
 using namespace std;
 using namespace gensios;
+#include <gensio/gensio_osops_env.h>
 
 // Internal gensio errors come in through this mechanism.
 class Logger: public Os_Funcs_Log_Handler {
@@ -25,55 +27,63 @@ class Logger: public Os_Funcs_Log_Handler {
     }
 };
 
-#ifdef _WIN32
-#define DEFAULT_ECHO_COMMPORT "COM0"
-bool
-file_is_accessible_dev(const char *filename)
-{
-    return true;
-}
-#else
+#define ECHO_DEV_ENV "GENSIO_TEST_ECHO_DEV"
 #define DEFAULT_ECHO_COMMPORT "/dev/ttyEcho0"
 
-bool
-file_is_accessible_dev(const char *filename)
+static int
+get_echo_dev(Os_Funcs &o, char *echo_dev, gensiods len)
 {
-    struct stat sb;
     int rv;
 
-    rv = stat(filename, &sb);
+    rv = gensio_os_env_get(ECHO_DEV_ENV, echo_dev, &len);
+#ifdef _WIN32
+    if (rv)
+	return rv;
+#else
+    if (rv == GE_NOTFOUND) {
+	if (len < strlen(DEFAULT_ECHO_COMMPORT) + 1)
+	    return GE_TOOBIG;
+	strncpy(echo_dev, DEFAULT_ECHO_COMMPORT, len);
+    } else if (rv) {
+	return rv;
+    }
+
+    struct stat sb;
+
+    rv = stat(echo_dev, &sb);
     if (rv == -1)
-	return false;
+	return gensio_os_err_to_err(o, errno);
 
     if (!S_ISCHR(sb.st_mode))
-	return false;
+	return GE_INVAL;
 
-    rv = open(filename, O_RDWR);
-    if (rv >= 0) {
-	close(rv);
-	return true;
-    } else {
-	return false;
-    }
-}
+    rv = open(echo_dev, O_RDWR);
+    if (rv < 0)
+	return gensio_os_err_to_err(o, errno);
+    close(rv);
 #endif
+    return 0;
+}
 
 int main(int argc, char *argv[])
 {
-    if (!file_is_accessible_dev(DEFAULT_ECHO_COMMPORT))
-	return 77;
-
     Os_Funcs o(0, new Logger);
+    char echo_dev[100];
     int err;
+
+    err = get_echo_dev(o, echo_dev, sizeof(echo_dev));
+    if (err) {
+	printf("Unable to get echo dev, test skipped: %s\n",
+	       gensio_err_to_str(err));
+	return 77;
+    }
+    std::string devstr(echo_dev);
+    devstr.append(",9600n81");
+
     Waiter w(o);
     static const char *serial_parms[] = { "nouucplock=false", NULL };
-#ifdef _WIN32
-    Gensio *bg = gensio_alloc("serialdev", "COM1,9600N81", serial_parms,
+    Gensio *bg = gensio_alloc("serialdev", devstr.c_str(), serial_parms,
 			      o, NULL);
-#else
-    Gensio *bg = gensio_alloc("serialdev", "/dev/ttyEcho0,9600n81",
-			      serial_parms, o, NULL);
-#endif
     Serial_Gensio *sg = (Serial_Gensio *) bg;
     GensioW g(sg); // Take over lifetime of the gensio
     unsigned int v;
