@@ -19,11 +19,14 @@ enum gensio_ll_close_state {
 struct gensio_iod;
 
 struct gensio_fd_ll_ops {
-    int (*sub_open)(void *handler_data, struct gensio_iod **iod);
+    int (*sub_open)(void *handler_data, struct gensio_iod **iod,
+		    gensio_time *next_timeout);
 
-    int (*check_open)(void *handler_data, struct gensio_iod *iod);
+    int (*check_open)(void *handler_data, struct gensio_iod *iod,
+		      gensio_time *next_timeout);
 
-    int (*retry_open)(void *handler_data, struct gensio_iod **iod);
+    int (*retry_open)(void *handler_data, struct gensio_iod **iod,
+		      gensio_time *next_timeout);
 
     /*
      * When GENSIO_LL_CLOSE_STATE_START, timeout will be NULL and the
@@ -55,6 +58,10 @@ struct gensio_fd_ll_ops {
     int (*acontrol)(void *handler_data, struct gensio_iod *iod, bool get,
 		    unsigned int option,
 		    struct gensio_func_acontrol *data);
+
+    gensiods (*deliver_read)(void *handler_data, void *cb_data, gensio_ll_cb cb, int err,
+			     void *data, gensiods datalen,
+			     const void *auxdata);
 };
 
 GENSIO_DLL_PUBLIC
