@@ -1461,7 +1461,9 @@ basen_set_read_callback_enable(struct basen_data *ndata, bool enabled)
 	/* Nothing to do, let the read/open handling wake things up. */
 	ndata->deferred_read = true;
     } else if (enabled && (read_pending || ndata->ll_err) &&
-	       ndata->state == BASEN_OPEN) {
+	       (ndata->state == BASEN_OPEN ||
+		ndata->state == BASEN_IO_ERR_CLOSE ||
+		ndata->state == BASEN_IN_LL_IO_ERR_CLOSE)) {
 	ndata->deferred_read = true;
 	basen_sched_deferred_op(ndata);
     } else {
@@ -1543,6 +1545,19 @@ gensio_base_func(struct gensio *io, int func, gensiods *count,
 	    return rv;
 	return rv2;
 
+    case GENSIO_FUNC_ACONTROL:
+	rv = GE_NOTSUP;
+	if (ndata->filter) {
+	    rv = gensio_filter_acontrol(ndata->filter, *((bool *) cbuf), buflen,
+					buf);
+	    if (rv && rv != GE_NOTSUP)
+		return rv;
+	}
+	rv2 = gensio_ll_acontrol(ndata->ll, *((bool *) cbuf), buflen, buf);
+	if (rv2 == GE_NOTSUP)
+	    return rv;
+	return rv2;
+
     case GENSIO_FUNC_DISABLE:
 	if (ndata->state != BASEN_CLOSED) {
 	    basen_set_state(ndata, BASEN_CLOSED);
@@ -2151,12 +2166,41 @@ gensio_filter_control(struct gensio_filter *filter, bool get,
 			NULL, data, datalen, NULL, &get, option, NULL);
 }
 
+int
+gensio_filter_acontrol(struct gensio_filter *filter, bool get,
+		       unsigned int option,
+		       struct gensio_func_acontrol *data)
+{
+    return filter->func(filter, GENSIO_FILTER_FUNC_ACONTROL,
+			NULL, data, 0, NULL, &get, option, NULL);
+}
+
 struct gensio *
 gensio_filter_get_gensio(struct gensio_filter *filter)
 {
     return filter->ndata->io;
 }
 
+void
+gensio_filter_vlog(struct gensio_filter *filter,
+		   enum gensio_log_levels level,
+		   const char *log, va_list args)
+{
+    gensio_gvlog(filter->ndata->io, level, log, args);
+}
+
+void
+gensio_filter_log(struct gensio_filter *filter,
+		  enum gensio_log_levels level,
+		  const char *log, ...)
+{
+    va_list args;
+
+    va_start(args, log);
+    gensio_filter_vlog(filter, level, log, args);
+    va_end(args);
+}
+
 int
 gensio_filter_do_event(struct gensio_filter *filter, int event, int err,
 		       unsigned char *buf, gensiods *buflen,
@@ -2260,6 +2304,14 @@ gensio_ll_control(struct gensio_ll *ll, bool get, int option, char *data,
 		    NULL);
 }
 
+int
+gensio_ll_acontrol(struct gensio_ll *ll, bool get, int option,
+		   struct gensio_func_acontrol *data)
+{
+    return ll->func(ll, GENSIO_LL_FUNC_ACONTROL, 0, data, &get,
+		    option, NULL);
+}
+
 int
 gensio_ll_do_event(struct gensio_ll *ll, int event, int err,
 		   unsigned char *buf, gensiods *buflen,
@@ -2304,3 +2356,23 @@ base_gensio_get_ll(struct gensio *io)
 
     return ndata->ll;
 }
+
+void
+gensio_ll_vlog(struct gensio_ll *ll,
+	       enum gensio_log_levels level,
+	       const char *log, va_list args)
+{
+    gensio_gvlog(ll->ndata->io, level, log, args);
+}
+
+void
+gensio_ll_log(struct gensio_ll *ll,
+	      enum gensio_log_levels level,
+	      const char *log, ...)
+{
+    va_list args;
+
+    va_start(args, log);
+    gensio_ll_vlog(ll, level, log, args);
+    va_end(args);
+}
