@@ -7,7 +7,13 @@
 
 /* This code handles TCP and Unix network I/O. */
 
+
 #include "config.h"
+#ifdef __linux__
+#define _GNU_SOURCE /* For struct ucred. */
+#define HAVE_NATIVE_UCRED
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -31,9 +37,62 @@
 #include <gensio/gensio_ll_fd.h>
 #include <gensio/argvutils.h>
 #include <gensio/gensio_osops.h>
+#include <gensio/gensio_time.h>
 
 #include "gensio_net.h"
 
+#ifdef _WIN32
+#include <windows.h>
+#include <afunix.h>
+#endif
+
+#ifdef HAVE_NATIVE_UCRED
+static int
+netna_get_ucred(struct gensio_os_funcs *o, int fd, struct ucred *cred)
+{
+    socklen_t len = sizeof(*cred);
+
+    if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, cred, &len) == -1)
+	return gensio_os_err_to_err(o, errno);
+    return 0;
+}
+#elif HAVE_UCRED
+struct ucred {
+    pid_t pid;
+    uid_t uid;
+    gid_t gid;
+};
+#endif
+
+/*
+ * The below is based on the code in glib2, gio/gcredentials.  If you
+ * want to implement this for something else, like other BSDs or
+ * solaris, you can see that code for what to do.
+ */
+#if defined(__APPLE__) || defined(__FreeBSD__)
+#include <sys/ucred.h>
+static int
+netna_get_ucred(struct gensio_os_funcs *o, int fd, struct ucred *cred)
+{
+    struct xucred c;
+    socklen_t len = sizeof(c);
+    pid_t pid = 0;
+
+    if (getsockopt(fd, SOL_LOCAL, LOCAL_PEERCRED, &c, &len) == -1)
+	return gensio_os_err_to_err(o, errno);
+#ifdef LOCAL_PEERPID
+    len = sizeof(pid);
+    if (getsockopt(fd, SOL_LOCAL, LOCAL_PEERPID, &pid, &len) == -1)
+	return gensio_os_err_to_err(o, errno);
+#endif
+
+    cred->pid = pid;
+    cred->uid = c.cr_uid;
+    cred->gid = c.cr_gid;
+    return 0;
+}
+#endif
+
 struct net_data {
     struct gensio_os_funcs *o;
 
@@ -44,7 +103,8 @@ struct net_data {
 
     bool nodelay;
 
-    bool istcp;
+    int protocol;
+    const char *typestr;
 
     int last_err;
 
@@ -52,7 +112,8 @@ struct net_data {
     int oob_char;
 };
 
-static int net_check_open(void *handler_data, struct gensio_iod *iod)
+static int net_check_open(void *handler_data, struct gensio_iod *iod,
+			  gensio_time *timeout)
 {
     struct net_data *tdata = handler_data;
 
@@ -62,23 +123,23 @@ static int net_check_open(void *handler_data, struct gensio_iod *iod)
 }
 
 static int
-net_try_open(struct net_data *tdata, struct gensio_iod **iod)
+net_try_open(struct net_data *tdata, struct gensio_iod **iod,
+	     gensio_time *timeout)
 {
     struct gensio_iod *new_iod = NULL;
     int err = GE_INUSE;
-    int protocol = tdata->istcp ? GENSIO_NET_PROTOCOL_TCP
-				: GENSIO_NET_PROTOCOL_UNIX;
-    unsigned int setup = (GENSIO_SET_OPENSOCK_REUSEADDR |
-			  GENSIO_OPENSOCK_REUSEADDR |
-			  GENSIO_SET_OPENSOCK_KEEPALIVE |
-			  GENSIO_SET_OPENSOCK_NODELAY);
+    unsigned int setup = 0;
 
-    if (tdata->istcp)
-	setup |= GENSIO_OPENSOCK_KEEPALIVE;
-    if (tdata->nodelay)
-	setup |= GENSIO_OPENSOCK_NODELAY;
+    if (tdata->protocol == GENSIO_NET_PROTOCOL_TCP) {
+        setup = (GENSIO_SET_OPENSOCK_REUSEADDR |
+		 GENSIO_OPENSOCK_REUSEADDR |
+		 GENSIO_SET_OPENSOCK_NODELAY);
+	setup |= GENSIO_SET_OPENSOCK_KEEPALIVE | GENSIO_OPENSOCK_KEEPALIVE;
+	if (tdata->nodelay)
+	    setup |= GENSIO_OPENSOCK_NODELAY;
+    }
  retry:
-    err = tdata->o->socket_open(tdata->o, tdata->ai, protocol, &new_iod);
+    err = tdata->o->socket_open(tdata->o, tdata->ai, tdata->protocol, &new_iod);
     if (err)
 	goto out;
 
@@ -118,22 +179,24 @@ net_try_open(struct net_data *tdata, struct gensio_iod **iod)
 }
 
 static int
-net_retry_open(void *handler_data, struct gensio_iod **iod)
+net_retry_open(void *handler_data, struct gensio_iod **iod,
+	       gensio_time *timeout)
 {
     struct net_data *tdata = handler_data;
 
     if (!gensio_addr_next(tdata->ai))
 	return tdata->last_err;
-    return net_try_open(tdata, iod);
+    return net_try_open(tdata, iod, timeout);
 }
 
 static int
-net_sub_open(void *handler_data, struct gensio_iod **iod)
+net_sub_open(void *handler_data, struct gensio_iod **iod,
+	     gensio_time *timeout)
 {
     struct net_data *tdata = handler_data;
 
     gensio_addr_rewind(tdata->ai);
-    return net_try_open(tdata, iod);
+    return net_try_open(tdata, iod, timeout);
 }
 
 static void
@@ -160,7 +223,7 @@ net_control(void *handler_data, struct gensio_iod *iod, bool get,
 
     switch (option) {
     case GENSIO_CONTROL_NODELAY:
-	if (!tdata->istcp)
+	if (tdata->protocol != GENSIO_NET_PROTOCOL_TCP)
 	    return GE_NOTSUP;
 	if (get) {
 	    if (iod) {
@@ -284,7 +347,7 @@ net_except_ready(void *handler_data, struct gensio_iod *iod)
     gensiods rcount = 0;
     int rv;
 
-    if (!tdata->istcp)
+    if (tdata->protocol != GENSIO_NET_PROTOCOL_TCP)
 	return GE_NOTSUP;
 
     rv = tdata->o->recv(iod, &urgdata, 1, &rcount, GENSIO_MSG_OOB);
@@ -352,9 +415,8 @@ static const struct gensio_fd_ll_ops net_fd_ll_ops = {
 
 static int
 net_gensio_alloc(const struct gensio_addr *iai, const char * const args[],
-		 struct gensio_os_funcs *o,
-		 gensio_event cb, void *user_data, const char *type,
-		 struct gensio **new_gensio)
+		 struct gensio_os_funcs *o, gensio_event cb, void *user_data,
+		 int protocol, const char *typestr, struct gensio **new_gensio)
 {
     struct net_data *tdata = NULL;
     struct gensio_addr *laddr = NULL, *laddr2, *addr = NULL;
@@ -364,24 +426,24 @@ net_gensio_alloc(const struct gensio_addr *iai, const char * const args[],
     unsigned int i;
     int ival;
     int err;
-    bool istcp = strcmp(type, "tcp") == 0;
-    GENSIO_DECLARE_PPGENSIO(p, o, cb, istcp ? "tcp" : "unix", user_data);
+    bool istcp = protocol == GENSIO_NET_PROTOCOL_TCP;
+    GENSIO_DECLARE_PPGENSIO(p, o, cb, typestr, user_data);
 
-    err = gensio_get_default(o, type, "nodelay", false,
+    err = gensio_get_default(o, typestr, "nodelay", false,
 			     GENSIO_DEFAULT_BOOL, NULL, &ival);
     if (err)
 	return err;
     nodelay = ival;
 
-    err = gensio_get_defaultaddr(o, type, "laddr", false,
+    err = gensio_get_defaultaddr(o, typestr, "laddr", false,
 				 GENSIO_NET_PROTOCOL_TCP, true, false, &laddr);
     if (err && err != GE_NOTSUP) {
 	gensio_log(o, GENSIO_LOG_ERR, "Invalid default %d laddr: %s",
-		   type, gensio_err_to_str(err));
+		   typestr, gensio_err_to_str(err));
 	return err;
     }
 
-    err = gensio_get_default(o, type, "nodelay", false,
+    err = gensio_get_default(o, typestr, "nodelay", false,
 			     GENSIO_DEFAULT_BOOL, NULL, &ival);
     if (err)
 	return err;
@@ -411,7 +473,8 @@ net_gensio_alloc(const struct gensio_addr *iai, const char * const args[],
     if (!tdata)
 	goto out_nomem;
 
-    tdata->istcp = istcp;
+    tdata->protocol = protocol;
+    tdata->typestr = typestr;
     tdata->oob_char = -1;
 
     addr = gensio_addr_dup(iai);
@@ -426,7 +489,7 @@ net_gensio_alloc(const struct gensio_addr *iai, const char * const args[],
     if (!tdata->ll)
 	goto out_nomem;
 
-    io = base_gensio_alloc(o, tdata->ll, NULL, NULL, type, cb, user_data);
+    io = base_gensio_alloc(o, tdata->ll, NULL, NULL, typestr, cb, user_data);
     if (!io)
 	goto out_nomem;
 
@@ -435,6 +498,8 @@ net_gensio_alloc(const struct gensio_addr *iai, const char * const args[],
     tdata->lai = laddr;
 
     gensio_set_is_reliable(io, true);
+    if (protocol == GENSIO_NET_PROTOCOL_UNIX_SEQPACKET)
+	gensio_set_is_packet(io, true);
 
     *new_gensio = io;
     return 0;
@@ -466,14 +531,14 @@ str_to_net_gensio(const char *str, const char * const args[],
 
     err = gensio_os_scan_netaddr(o, str, false, protocol, &addr);
     if (err) {
-	bool istcp = strcmp(typestr, "tcp") == 0;
-	GENSIO_DECLARE_PPGENSIO(p, o, cb, istcp ? "tcp" : "unix", user_data);
+	GENSIO_DECLARE_PPGENSIO(p, o, cb, typestr, user_data);
 
 	gensio_pparm_log(&p, "Invalid network address: %s", str);
 	return err;
     }
 
-    err = net_gensio_alloc(addr, args, o, cb, user_data, typestr, new_gensio);
+    err = net_gensio_alloc(addr, args, o, cb, user_data, protocol,
+			   typestr, new_gensio);
     gensio_addr_free(addr);
 
     return err;
@@ -487,7 +552,8 @@ tcp_gensio_alloc(const void *gdata, const char * const args[],
 {
     const struct gensio_addr *iai = gdata;
 
-    return net_gensio_alloc(iai, args, o, cb, user_data, "tcp", new_gensio);
+    return net_gensio_alloc(iai, args, o, cb, user_data,
+			    GENSIO_NET_PROTOCOL_TCP, "tcp", new_gensio);
 }
 
 static int
@@ -509,7 +575,25 @@ unix_gensio_alloc(const void *gdata, const char * const args[],
 #if HAVE_UNIX
     const struct gensio_addr *iai = gdata;
 
-    return net_gensio_alloc(iai, args, o, cb, user_data, "unix", new_gensio);
+    return net_gensio_alloc(iai, args, o, cb, user_data,
+			    GENSIO_NET_PROTOCOL_UNIX, "unix", new_gensio);
+#else
+    return GE_NOTSUP;
+#endif
+}
+
+static int
+unixseq_gensio_alloc(const void *gdata, const char * const args[],
+		     struct gensio_os_funcs *o,
+		     gensio_event cb, void *user_data,
+		     struct gensio **new_gensio)
+{
+#if HAVE_UNIX && defined(SOCK_SEQPACKET)
+    const struct gensio_addr *iai = gdata;
+
+    return net_gensio_alloc(iai, args, o, cb, user_data,
+			    GENSIO_NET_PROTOCOL_UNIX_SEQPACKET, "unixseq",
+			    new_gensio);
 #else
     return GE_NOTSUP;
 #endif
@@ -525,6 +609,20 @@ str_to_unix_gensio(const char *str, const char * const args[],
 			     o, cb, user_data, new_gensio);
 }
 
+static int
+str_to_unixseq_gensio(const char *str, const char * const args[],
+		      struct gensio_os_funcs *o,
+		      gensio_event cb, void *user_data,
+		      struct gensio **new_gensio)
+{
+#if HAVE_UNIX && defined(SOCK_SEQPACKET)
+    return str_to_net_gensio(str, args, GENSIO_NET_PROTOCOL_UNIX_SEQPACKET,
+			     "unixseq", o, cb, user_data, new_gensio);
+#else
+    return GE_NOTSUP;
+#endif
+}
+
 struct netna_data {
     struct gensio_accepter *acc;
 
@@ -549,6 +647,10 @@ struct netna_data {
     bool mode_set;
     char *owner;
     char *group;
+#if HAVE_UCRED
+    const char **permusers; /* Permitted user list. */
+    const char **permgrps; /* Permitted group list. */
+#endif
 #endif
 
 #ifdef HAVE_TCPD_H
@@ -561,7 +663,8 @@ struct netna_data {
 
     unsigned int opensock_flags;
 
-    bool istcp;
+    int protocol;
+    const char *typestr;
 };
 
 static const struct gensio_fd_ll_ops net_server_fd_ll_ops = {
@@ -614,6 +717,42 @@ netna_finish_server_open(struct gensio *net, int err, void *cb_data)
     base_gensio_server_open_done(nadata->acc, net, err);
 }
 
+#if HAVE_UCRED
+static bool
+netna_permitted(struct ucred *cred, struct netna_data *nadata)
+{
+    unsigned int i;
+    char buf[16384];
+    int err;
+
+    /*
+     * Either permusers or permgroups must be set here, so we don't
+     * have to check for both being NULL.
+     */
+    for (i = 0; nadata->permusers && nadata->permusers[i]; i++) {
+	struct passwd pwdbuf, *pwd;
+
+	err = getpwnam_r(nadata->permusers[i], &pwdbuf, buf, sizeof(buf),
+			 &pwd);
+	if (err || !pwd)
+	    continue;
+	if (cred->uid == pwd->pw_uid)
+	    return true;
+    }
+    for (i = 0; nadata->permgrps && nadata->permgrps[i]; i++) {
+	struct group grpbuf, *grp;
+
+	err = getgrnam_r(nadata->permgrps[i], &grpbuf, buf, sizeof(buf),
+			 &grp);
+	if (err || !grp)
+	    continue;
+	if (cred->gid == grp->gr_gid)
+	    return true;
+    }
+    return false;
+}
+#endif
+
 static void
 netna_readhandler(struct gensio_iod *iod, void *cbdata)
 {
@@ -622,12 +761,15 @@ netna_readhandler(struct gensio_iod *iod, void *cbdata)
     struct gensio_addr *raddr;
     struct net_data *tdata = NULL;
     struct gensio *io = NULL;
-    unsigned int setup = (GENSIO_SET_OPENSOCK_REUSEADDR |
-			  GENSIO_OPENSOCK_REUSEADDR |
-			  GENSIO_SET_OPENSOCK_KEEPALIVE |
-			  GENSIO_SET_OPENSOCK_NODELAY);
+    unsigned int setup = 0;
     int err;
 
+    if (nadata->protocol == GENSIO_NET_PROTOCOL_TCP) {
+	setup = (GENSIO_SET_OPENSOCK_REUSEADDR |
+		 GENSIO_OPENSOCK_REUSEADDR |
+		 GENSIO_SET_OPENSOCK_KEEPALIVE |
+		 GENSIO_SET_OPENSOCK_NODELAY);
+    }
     err = nadata->o->accept(iod, &raddr, &new_iod);
     if (err) {
 	if (err != GE_NODATA)
@@ -646,7 +788,8 @@ netna_readhandler(struct gensio_iod *iod, void *cbdata)
     }
 
 #ifdef HAVE_TCPD_H
-    if (nadata->istcp && nadata->tcpd != GENSIO_TCPD_OFF) {
+    if (nadata->protocol == GENSIO_NET_PROTOCOL_TCP &&
+		nadata->tcpd != GENSIO_TCPD_OFF) {
 	const char *msg = gensio_os_check_tcpd_ok(new_iod,
 						  nadata->tcpd_progname);
 
@@ -662,6 +805,28 @@ netna_readhandler(struct gensio_iod *iod, void *cbdata)
 	}
     }
 #endif
+#if HAVE_UCRED
+    if (nadata->protocol != GENSIO_NET_PROTOCOL_TCP &&
+		(nadata->permusers || nadata->permgrps)) {
+	int fd = nadata->o->iod_get_fd(new_iod);
+	struct ucred cred;
+
+	err = netna_get_ucred(nadata->o, fd, &cred);
+	if (err) {
+	    gensio_acc_log(nadata->acc, GENSIO_LOG_INFO,
+			   "Error getting peer credentials: %s",
+			   gensio_err_to_str(err));
+	    goto out_err;
+	}
+
+	if (!netna_permitted(&cred, nadata)) {
+	    gensio_acc_log(nadata->acc, GENSIO_LOG_INFO,
+			   "Error accepting unix gensio: user not permitted");
+	    err = GE_PERM;
+	    goto out_err;
+	}
+    }
+#endif
 
     tdata = nadata->o->zalloc(nadata->o, sizeof(*tdata));
     if (!tdata) {
@@ -674,11 +839,11 @@ netna_readhandler(struct gensio_iod *iod, void *cbdata)
     tdata->o = nadata->o;
     tdata->oob_char = -1;
     tdata->ai = raddr;
-    tdata->istcp = nadata->istcp;
+    tdata->protocol = nadata->protocol;
     tdata->nodelay = nadata->nodelay;
     raddr = NULL;
 
-    if (tdata->istcp)
+    if (tdata->protocol == GENSIO_NET_PROTOCOL_TCP)
 	setup |= GENSIO_OPENSOCK_KEEPALIVE;
     if (tdata->nodelay)
 	setup |= GENSIO_OPENSOCK_NODELAY;
@@ -699,7 +864,7 @@ netna_readhandler(struct gensio_iod *iod, void *cbdata)
     }
 
     io = base_gensio_server_alloc(nadata->o, tdata->ll, NULL, NULL,
-				  nadata->istcp ? "tcp" : "unix",
+				  nadata->typestr,
 				  netna_finish_server_open, nadata);
     if (!io) {
 	gensio_acc_log(nadata->acc, GENSIO_LOG_ERR,
@@ -708,6 +873,8 @@ netna_readhandler(struct gensio_iod *iod, void *cbdata)
 	goto out_err;
     }
     gensio_set_is_reliable(io, true);
+    if (tdata->protocol == GENSIO_NET_PROTOCOL_UNIX_SEQPACKET)
+	gensio_set_is_packet(io, true);
     err = base_gensio_server_start(io);
     if (err)
 	goto out_err;
@@ -735,7 +902,16 @@ netna_readhandler(struct gensio_iod *iod, void *cbdata)
 	nadata->o->close(&new_iod);
 }
 
-#if HAVE_UNIX
+#if HAVE_UNIX || defined(_WIN32)
+/*
+ * This section is duplicated in gensio_dgram.c.  If you fix something
+ * here, fix it there, too.
+ */
+
+#define SIZEOF_SOCKADDR_UN_HEADER \
+    (sizeof(struct sockaddr_un) - \
+     sizeof(((struct sockaddr_un *) 0)->sun_path))
+
 #define MAX_UNIX_ADDR_PATH (sizeof(((struct sockaddr_un *) 0)->sun_path) + 1)
 static void
 get_unix_addr_path(struct gensio_addr *addr, char *path)
@@ -746,28 +922,35 @@ get_unix_addr_path(struct gensio_addr *addr, char *path)
 
     /* Remove the socket if it already exists. */
     gensio_addr_getaddr(addr, sun, &len);
+    len -= SIZEOF_SOCKADDR_UN_HEADER;
 
     /*
      * Make sure the path is nil terminated.  See discussions
      * in the unix(7) man page on Linux for details.
      */
-    memcpy(path, sun->sun_path, len - sizeof(sa_family_t));
-    path[len - sizeof(sa_family_t)] = '\0';
+    memcpy(path, sun->sun_path, len);
+    path[len] = '\0';
 }
 #endif
 
 static void
 netna_rm_unix_socket(struct gensio_addr *addr)
 {
-#if HAVE_UNIX
+#if HAVE_UNIX || defined(_WIN32)
     char path[MAX_UNIX_ADDR_PATH];
 
     /* Remove the socket if it already exists. */
     get_unix_addr_path(addr, path);
+#endif
+#if HAVE_UNIX
     unlink(path);
 #endif
+#ifdef _WIN32
+    DeleteFile(path);
+#endif
 }
 
+/* Duplicated in gensio_dgram.c */
 static int
 netna_b4_listen(struct gensio_iod *iod, void *data)
 {
@@ -782,7 +965,7 @@ netna_b4_listen(struct gensio_iod *iod, void *data)
     char unpath[MAX_UNIX_ADDR_PATH];
 #endif
 
-    if (nadata->istcp)
+    if (nadata->protocol == GENSIO_NET_PROTOCOL_TCP)
 	return 0;
 
 #if HAVE_UNIX
@@ -811,7 +994,6 @@ netna_b4_listen(struct gensio_iod *iod, void *data)
     if (nadata->group) {
 	struct group grpbuf, *grp;
 
-
 	err = getgrnam_r(nadata->group, &grpbuf, pwbuf, sizeof(pwbuf), &grp);
 	if (err)
 	    goto out_errno;
@@ -835,7 +1017,7 @@ netna_b4_listen(struct gensio_iod *iod, void *data)
  out_err:
     return gensio_os_err_to_err(nadata->o, err);
 #else
-    return GE_NOTSUP;
+    return 0;
 #endif
 }
 
@@ -866,7 +1048,7 @@ netna_shutdown(struct gensio_accepter *accepter,
     for (i = 0; i < nadata->nr_acceptfds; i++)
 	nadata->o->clear_fd_handlers(nadata->acceptfds[i].iod);
 
-    if (!nadata->istcp)
+    if (nadata->protocol != GENSIO_NET_PROTOCOL_TCP)
 	/* Remove the socket. */
 	netna_rm_unix_socket(nadata->ai);
 
@@ -917,6 +1099,12 @@ netna_free(struct gensio_accepter *accepter, struct netna_data *nadata)
 	nadata->o->free(nadata->o, nadata->owner);
     if (nadata->group)
 	nadata->o->free(nadata->o, nadata->group);
+#if HAVE_UCRED
+    if (nadata->permusers)
+	gensio_argv_free(nadata->o, nadata->permusers);
+    if (nadata->permgrps)
+	gensio_argv_free(nadata->o, nadata->permgrps);
+#endif
 #endif
 #ifdef HAVE_TCPD_H
     if (nadata->tcpd_progname)
@@ -942,7 +1130,7 @@ netna_str_to_gensio(struct gensio_accepter *accepter,
     int protocol = 0;
     bool nodelay = false;
     GENSIO_DECLARE_PPGENSIO(p, nadata->o, cb,
-			    nadata->istcp ? "tcp" : "unix", user_data);
+			    nadata->typestr, user_data);
 
     err = gensio_scan_network_port(nadata->o, addr, false, &ai,
 				   &protocol, &is_port_set, NULL, &iargs);
@@ -952,23 +1140,20 @@ netna_str_to_gensio(struct gensio_accepter *accepter,
     }
 
     err = GE_INVAL;
-    if (nadata->istcp) {
-	if (protocol != GENSIO_NET_PROTOCOL_TCP || !is_port_set)
-	    goto out_err;
-    } else {
-	if (protocol != GENSIO_NET_PROTOCOL_UNIX)
-	    goto out_err;
-    }
+    if (nadata->protocol != protocol)
+	goto out_err;
+    if (protocol == GENSIO_NET_PROTOCOL_TCP && !is_port_set)
+	goto out_err;
 
     for (i = 0; iargs && iargs[i]; i++) {
 	if (gensio_pparm_ds(&p, iargs[i], "readbuf", &max_read_size) > 0)
 	    continue;
-	if (nadata->istcp &&
+	if (nadata->protocol == GENSIO_NET_PROTOCOL_TCP &&
 		gensio_pparm_value(&p, iargs[i], "laddr", &dummy) > 0) {
 	    laddr = iargs[i];
 	    continue;
 	}
-	if (nadata->istcp &&
+	if (nadata->protocol == GENSIO_NET_PROTOCOL_TCP &&
 		gensio_pparm_bool(&p, args[i], "nodelay", &nodelay) > 0)
 	    continue;
 	gensio_pparm_unknown_parm(&p, args[i]);
@@ -989,7 +1174,7 @@ netna_str_to_gensio(struct gensio_accepter *accepter,
 	args[i++] = "nodelay";
 
     err = net_gensio_alloc(ai, args, nadata->o, cb, user_data,
-			   nadata->istcp ? "tcp" : "unix", new_io);
+			   nadata->protocol, nadata->typestr, new_io);
 
  out_err:
     if (iargs)
@@ -1042,7 +1227,7 @@ netna_control_lport(struct netna_data *nadata, bool get,
     if (!get)
 	return GE_NOTSUP;
 
-    if (!nadata->istcp)
+    if (nadata->protocol != GENSIO_NET_PROTOCOL_TCP)
 	return GE_NOTSUP;
 
     if (nadata->nr_acceptfds == 0)
@@ -1151,34 +1336,38 @@ net_gensio_accepter_alloc(const struct gensio_addr *iai,
 			  const char * const args[],
 			  struct gensio_os_funcs *o,
 			  gensio_accepter_event cb, void *user_data,
-			  const char *type,
+			  int protocol, const char *typestr,
 			  struct gensio_accepter **accepter)
 {
     struct netna_data *nadata;
     gensiods max_read_size = GENSIO_DEFAULT_BUF_SIZE;
     bool nodelay = false;
-    bool istcp = strcmp(type, "tcp") == 0;
-    bool reuseaddr = istcp ? true : false;
+    bool reuseaddr = protocol == GENSIO_NET_PROTOCOL_TCP;
 #if HAVE_UNIX
     unsigned int umode = 6, gmode = 6, omode = 6, mode;
     bool mode_set = false;
     const char *owner = NULL, *group = NULL;
+#if HAVE_UCRED
+    const char **permusers = NULL;
+    const char **permgrps = NULL;
+#endif
 #endif
 #ifdef HAVE_TCPD_H
     const char *tcpdname = NULL;
     enum gensio_tcpd_options tcpd = GENSIO_TCPD_ON;
 #endif
+    bool istcp = protocol == GENSIO_NET_PROTOCOL_TCP;
     unsigned int i;
     int err, ival;
-    GENSIO_DECLARE_PPACCEPTER(p, o, cb, istcp ? "tcp" : "unix", user_data);
+    GENSIO_DECLARE_PPACCEPTER(p, o, cb, typestr, user_data);
 
-    if (istcp) {
-	err = gensio_get_default(o, type, "reuseaddr", false,
+    if (protocol == GENSIO_NET_PROTOCOL_TCP) {
+	err = gensio_get_default(o, typestr, "reuseaddr", false,
 				 GENSIO_DEFAULT_BOOL, NULL, &ival);
 	if (err)
 	    return err;
     } else {
-	err = gensio_get_default(o, type, "delsock", false,
+	err = gensio_get_default(o, typestr, "delsock", false,
 				 GENSIO_DEFAULT_BOOL, NULL, &ival);
 	if (err)
 	    return err;
@@ -1186,7 +1375,7 @@ net_gensio_accepter_alloc(const struct gensio_addr *iai,
     reuseaddr = ival;
 
 #ifdef HAVE_TCPD_H
-    err = gensio_get_default(o, type, "tcpd", false,
+    err = gensio_get_default(o, typestr, "tcpd", false,
 			     GENSIO_DEFAULT_INT, NULL, &ival);
     if (err)
 	return err;
@@ -1237,20 +1426,37 @@ net_gensio_accepter_alloc(const struct gensio_addr *iai,
 	    continue;
 	if (!istcp && gensio_pparm_value(&p, args[i], "group", &group))
 	    continue;
+#if HAVE_UCRED
+	if (!istcp && gensio_pparm_argv(&p, args[i], "permusers", ";",
+					NULL, &permusers))
+	    continue;
+	if (!istcp && gensio_pparm_argv(&p, args[i], "permgrps", ";",
+					NULL, &permgrps))
+	    continue;
+#endif
 #endif
 	gensio_pparm_unknown_parm(&p, args[i]);
-	return GE_INVAL;
+	err = GE_INVAL;
+	goto out_err2;
     }
 
     nadata = o->zalloc(o, sizeof(*nadata));
-    if (!nadata)
-	return GE_NOMEM;
+    if (!nadata) {
+	err = GE_NOMEM;
+	goto out_err2;
+    }
     nadata->o = o;
 
     err = GE_NOMEM;
     if (reuseaddr)
 	nadata->opensock_flags |= GENSIO_OPENSOCK_REUSEADDR;
 #if HAVE_UNIX
+#if HAVE_UCRED
+    nadata->permusers = permusers;
+    permusers = NULL;
+    nadata->permgrps = permgrps;
+    permgrps = NULL;
+#endif
     nadata->mode_set = mode_set;
     nadata->mode = umode << 6 | gmode << 3 | omode;
     if (owner) {
@@ -1283,15 +1489,18 @@ net_gensio_accepter_alloc(const struct gensio_addr *iai,
     if (!nadata->cb_en_done_runner)
 	goto out_err;
 
-    nadata->istcp = istcp;
+    nadata->protocol = protocol;
+    nadata->typestr = typestr;
 
     err = base_gensio_accepter_alloc(NULL, netna_base_acc_op, nadata,
-				    o, type, cb, user_data, accepter);
+				    o, typestr, cb, user_data, accepter);
     if (err)
 	goto out_err;
 
     nadata->acc = *accepter;
     gensio_acc_set_is_reliable(nadata->acc, true);
+    if (protocol == GENSIO_NET_PROTOCOL_UNIX_SEQPACKET)
+	gensio_acc_set_is_packet(nadata->acc, true);
     nadata->max_read_size = max_read_size;
     nadata->nodelay = nodelay;
 
@@ -1299,6 +1508,13 @@ net_gensio_accepter_alloc(const struct gensio_addr *iai,
 
  out_err:
     netna_free(NULL, nadata);
+ out_err2:
+#if HAVE_UCRED
+    if (permusers)
+	gensio_argv_free(o, permusers);
+    if (permgrps)
+	gensio_argv_free(o, permgrps);
+#endif
     return err;
 }
 
@@ -1315,14 +1531,14 @@ str_to_net_gensio_accepter(const char *str, const char * const args[],
 
     err = gensio_os_scan_netaddr(o, str, true, protocol, &ai);
     if (err) {
-	bool istcp = strcmp(typestr, "tcp") == 0;
-	GENSIO_DECLARE_PPACCEPTER(p, o, cb, istcp ? "tcp" : "unix", user_data);
+	GENSIO_DECLARE_PPACCEPTER(p, o, cb, typestr, user_data);
 
 	gensio_pparm_log(&p, "Invalid network address: %s", str);
 	return err;
     }
 
-    err = net_gensio_accepter_alloc(ai, args, o, cb, user_data, typestr, acc);
+    err = net_gensio_accepter_alloc(ai, args, o, cb, user_data,
+				    protocol, typestr, acc);
     gensio_addr_free(ai);
 
     return err;
@@ -1337,7 +1553,8 @@ tcp_gensio_accepter_alloc(const void *gdata,
 {
     const struct gensio_addr *iai = gdata;
 
-    return net_gensio_accepter_alloc(iai, args, o, cb, user_data, "tcp",
+    return net_gensio_accepter_alloc(iai, args, o, cb, user_data,
+				     GENSIO_NET_PROTOCOL_TCP, "tcp",
 				     accepter);
 }
 
@@ -1359,10 +1576,26 @@ unix_gensio_accepter_alloc(const void *gdata,
 			   gensio_accepter_event cb, void *user_data,
 			   struct gensio_accepter **accepter)
 {
-#if HAVE_UNIX
     const struct gensio_addr *iai = gdata;
 
-    return net_gensio_accepter_alloc(iai, args, o, cb, user_data, "unix",
+    return net_gensio_accepter_alloc(iai, args, o, cb, user_data,
+				     GENSIO_NET_PROTOCOL_UNIX, "unix",
+				     accepter);
+}
+
+static int
+unixseq_gensio_accepter_alloc(const void *gdata,
+			      const char * const args[],
+			      struct gensio_os_funcs *o,
+			      gensio_accepter_event cb, void *user_data,
+			      struct gensio_accepter **accepter)
+{
+#if HAVE_UNIX && defined(SOCK_SEQPACKET)
+    const struct gensio_addr *iai = gdata;
+
+    return net_gensio_accepter_alloc(iai, args, o, cb, user_data,
+				     GENSIO_NET_PROTOCOL_UNIX_SEQPACKET,
+				     "unixseq",
 				     accepter);
 #else
     return GE_NOTSUP;
@@ -1380,6 +1613,22 @@ str_to_unix_gensio_accepter(const char *str, const char * const args[],
 				      "unix", o, cb, user_data, acc);
 }
 
+static int
+str_to_unixseq_gensio_accepter(const char *str, const char * const args[],
+			       struct gensio_os_funcs *o,
+			       gensio_accepter_event cb,
+			       void *user_data,
+			       struct gensio_accepter **acc)
+{
+#if HAVE_UNIX && defined(SOCK_SEQPACKET)
+    return str_to_net_gensio_accepter(str, args,
+				      GENSIO_NET_PROTOCOL_UNIX_SEQPACKET,
+				      "unixseq", o, cb, user_data, acc);
+#else
+    return GE_NOTSUP;
+#endif
+}
+
 int
 gensio_init_net(struct gensio_os_funcs *o)
 {
@@ -1399,5 +1648,13 @@ gensio_init_net(struct gensio_os_funcs *o)
 				  unix_gensio_accepter_alloc);
     if (rv)
 	return rv;
+    rv = register_gensio(o, "unixseq", str_to_unixseq_gensio,
+			 unixseq_gensio_alloc);
+    if (rv)
+	return rv;
+    rv = register_gensio_accepter(o, "unixseq", str_to_unixseq_gensio_accepter,
+				  unixseq_gensio_accepter_alloc);
+    if (rv)
+	return rv;
     return 0;
 }
