@@ -94,6 +94,12 @@ struct termio_op_q {
     struct termio_op_q *next;
 };
 
+struct modemstate_cb {
+    struct modemstate_cb *next;
+    gensio_control_done cdone;
+    void *cb_data;
+};
+
 struct sterm_data {
     struct gensio *io;
     struct sergensio *sio;
@@ -161,6 +167,9 @@ struct sterm_data {
     unsigned int modemstate_mask;
     bool handling_modemstate;
     bool sent_first_modemstate;
+    struct modemstate_cb *modemstate_cbs;
+
+    unsigned int linestate_mask;
 };
 
 static int
@@ -322,7 +331,8 @@ serconf_set_get(struct sterm_data *sdata, int op, int val, const char *sval,
 		const struct termio_xlat_str *xlatstr,
 		void (*done)(struct sergensio *sio, int err,
 			     int val, void *cb_data),
-		void *cb_data)
+		void *cb_data,
+		void (*finish)(struct sterm_data *sdata, int val))
 {
     struct termio_op_q *qe = NULL;
     int err = 0;
@@ -377,6 +387,8 @@ serconf_set_get(struct sterm_data *sdata, int op, int val, const char *sval,
 				    false, 0);
 	if (err)
 	    goto out_unlock;
+	if (finish)
+	    finish(sdata, val);
     }
 
     if (qe) {
@@ -407,7 +419,7 @@ sterm_baud(struct sterm_data *sdata, int baud, const char *sbaud,
 {
     return serconf_set_get(sdata,
 			   GENSIO_IOD_CONTROL_BAUD, baud, sbaud,
-			   NULL, cdone, NULL, done, cb_data);
+			   NULL, cdone, NULL, done, cb_data, NULL);
 }
 
 static int
@@ -419,7 +431,7 @@ sterm_datasize(struct sterm_data *sdata, int datasize, const char *sdatasize,
 {
     return serconf_set_get(sdata,
 			   GENSIO_IOD_CONTROL_DATASIZE, datasize, sdatasize,
-			   NULL, cdone, NULL, done, cb_data);
+			   NULL, cdone, NULL, done, cb_data, NULL);
 }
 
 static const struct termio_xlat_str sterm_parity_xlatstr[] = {
@@ -442,7 +454,8 @@ sterm_parity(struct sterm_data *sdata, int parity, const char *sparity,
 {
     return serconf_set_get(sdata,
 			   GENSIO_IOD_CONTROL_PARITY, parity, sparity,
-			   NULL, cdone, sterm_parity_xlatstr, done, cb_data);
+			   NULL, cdone, sterm_parity_xlatstr, done, cb_data,
+			   NULL);
 }
 
 static int
@@ -454,7 +467,7 @@ sterm_stopbits(struct sterm_data *sdata, int stopbits, const char *sstopbits,
 {
     return serconf_set_get(sdata,
 			   GENSIO_IOD_CONTROL_STOPBITS, stopbits, sstopbits,
-			   NULL, cdone, NULL, done, cb_data);
+			   NULL, cdone, NULL, done, cb_data, NULL);
 }
 
 static int
@@ -567,7 +580,7 @@ sterm_flowcontrol(struct sterm_data *sdata, int flowcontrol,
 
     return serconf_set_get(sdata, GENSIO_IOD_CONTROL_RTSCTS, flowcontrol,
 			   NULL, serconf_xlat_flowcontrol, cdone, xlatstr,
-			   done, cb_data);
+			   done, cb_data, NULL);
 }
 
 static const struct termio_xlat_str sterm_iflow_xlatstr[] = {
@@ -592,7 +605,7 @@ sterm_iflowcontrol(struct sterm_data *sdata, int iflowcontrol,
     return serconf_set_get(sdata,
 			   GENSIO_IOD_CONTROL_XONXOFF, 0, siflowcontrol,
 			   serconf_xlat_flowcontrol,
-			   cdone, sterm_iflow_xlatstr, done, cb_data);
+			   cdone, sterm_iflow_xlatstr, done, cb_data, NULL);
 }
 
 static const struct termio_xlat_str sterm_on_off_xlatstr[] = {
@@ -633,7 +646,7 @@ sterm_sbreak(struct sterm_data *sdata, int breakv, const char *sbreakv,
     return serconf_set_get(sdata,
 			   GENSIO_IOD_CONTROL_SET_BREAK, breakv, sbreakv,
 			   sterm_xlat_sbreak, cdone, sterm_on_off_xlatstr,
-			   done, cb_data);
+			   done, cb_data, NULL);
 }
 
 static int
@@ -666,7 +679,7 @@ sterm_dtr(struct sterm_data *sdata, int dtr, const char *sdtr,
     return serconf_set_get(sdata,
 			   GENSIO_IOD_CONTROL_DTR, dtr, sdtr,
 			   serconf_xlat_dtr, cdone, sterm_on_off_xlatstr,
-			   done, cb_data);
+			   done, cb_data, NULL);
 }
 
 static int
@@ -699,7 +712,7 @@ sterm_rts(struct sterm_data *sdata, int rts, const char *srts,
     return serconf_set_get(sdata,
 			   GENSIO_IOD_CONTROL_RTS, rts, srts,
 			   serconf_xlat_rts, cdone, sterm_on_off_xlatstr,
-			   done, cb_data);
+			   done, cb_data, NULL);
 }
 
 static void
@@ -708,6 +721,7 @@ serialdev_timeout(struct gensio_timer *t, void *cb_data)
     struct sterm_data *sdata = cb_data;
     int modemstate = 0, rv;
     bool force_send;
+    struct modemstate_cb *c, *n;
 
     sterm_lock(sdata);
     if (sdata->handling_modemstate || !sdata->open) {
@@ -728,8 +742,18 @@ serialdev_timeout(struct gensio_timer *t, void *cb_data)
     sdata->last_modemstate = modemstate & sdata->modemstate_mask;
     force_send = !sdata->sent_first_modemstate;
     sdata->sent_first_modemstate = true;
+    c = sdata->modemstate_cbs;
+    sdata->modemstate_cbs = NULL;
     sterm_unlock(sdata);
 
+    /* Call any modemstate callback operations. */
+    while (c) {
+	n = c->next;
+	c->cdone(sdata->io, 0, NULL, 0, c->cb_data);
+	sdata->o->free(sdata->o, c);
+	c = n;
+    }
+
     /*
      * The bottom 4 buts of modemstate is the "changed" bits, only
      * report this if someing changed that was in the mask.
@@ -755,16 +779,36 @@ serialdev_timeout(struct gensio_timer *t, void *cb_data)
 }
 
 static int
-sterm_modemstate(struct sterm_data *sdata, unsigned int val, const char *sval)
+sterm_modemstate_mask(struct sterm_data *sdata, unsigned int val,
+		      const char *sval,
+		      gensio_control_done cdone, void *cb_data)
 {
     gensio_time timeout = {0, 0};
+    struct modemstate_cb *c, *n;
 
     if (sval)
 	val = strtol(sval, NULL, 0);
 
+    if (cdone) {
+	n = sdata->o->zalloc(sdata->o, sizeof(*n));
+	if (!n)
+	    return GE_NOMEM;
+	n->cdone = cdone;
+	n->cb_data = cb_data;
+    }
+
     sterm_lock(sdata);
     sdata->modemstate_mask = val;
     sdata->sent_first_modemstate = false;
+    if (cdone) {
+	if (!sdata->modemstate_cbs) {
+	    sdata->modemstate_cbs = n;
+	} else {
+	    for (c = sdata->modemstate_cbs; c->next != NULL; c = c->next)
+		;
+	    c->next = n;
+	}
+    }
     sterm_unlock(sdata);
 
     /* Cause an immediate send of the modemstate. */
@@ -773,6 +817,27 @@ sterm_modemstate(struct sterm_data *sdata, unsigned int val, const char *sval)
     return 0;
 }
 
+static void
+sterm_finish_linestate(struct sterm_data *sdata, int val)
+{
+    sdata->linestate_mask = val;
+}
+
+static int
+sterm_linestate_mask(struct sterm_data *sdata, unsigned int val,
+		     const char *sval,
+		     gensio_control_done cdone, void *cb_data)
+{
+    if (sval)
+	val = strtol(sval, NULL, 0);
+
+    val &= GENSIO_SER_LINESTATE_PARITY_ERR | GENSIO_SER_LINESTATE_BREAK;
+
+    return serconf_set_get(sdata, GENSIO_IOD_CONTROL_ENABLE_RECV_ERR,
+			   val, NULL, NULL,
+			   cdone, NULL, NULL, cb_data, sterm_finish_linestate);
+}
+
 static int
 sterm_flowcontrol_state(struct sterm_data *sdata, bool val, char *sval)
 {
@@ -867,7 +932,7 @@ sergensio_sterm_func(struct sergensio *sio, int op, int val, char *buf,
     case SERGENSIO_FUNC_MODEMSTATE:
 	if (done)
 	    return GE_INVAL;
-	return sterm_modemstate(sdata, val, NULL);
+	return sterm_modemstate_mask(sdata, val, NULL, NULL, NULL);
 
     case SERGENSIO_FUNC_FLOWCONTROL_STATE:
 	if (done)
@@ -884,8 +949,15 @@ sergensio_sterm_func(struct sergensio *sio, int op, int val, char *buf,
 	    return GE_INVAL;
 	return sterm_send_break(sdata);
 
-    case SERGENSIO_FUNC_SIGNATURE:
     case SERGENSIO_FUNC_LINESTATE:
+	if (done)
+	    return GE_INVAL;
+	if (!sdata->o->read_flags)
+	    return GE_NOTSUP;
+
+	return sterm_linestate_mask(sdata, val, NULL, NULL, NULL);
+
+    case SERGENSIO_FUNC_SIGNATURE:
     default:
 	return GE_NOTSUP;
     }
@@ -935,6 +1007,14 @@ sterm_acontrol(void *handler_data, struct gensio_iod *iod, bool get,
     case GENSIO_ACONTROL_SER_RTS:
 	return sterm_rts(sdata, 0, data, done, NULL, cb_data);
 
+    case GENSIO_ACONTROL_SER_SET_MODEMSTATE_MASK:
+	return sterm_modemstate_mask(sdata, 0, data, done, cb_data);
+
+    case GENSIO_ACONTROL_SER_SET_LINESTATE_MASK:
+	if (!sdata->o->read_flags)
+	    return GE_NOTSUP;
+	return sterm_linestate_mask(sdata, 0, data, done, cb_data);
+
     default:
 	return GE_NOTSUP;
     }
@@ -1038,6 +1118,7 @@ is_a_pty(const char *ttyname)
 #else
     char buf[PATH_MAX];
 
+    memset(buf, 0, sizeof(buf));
     while (readlink(ttyname, buf, sizeof(buf)) > 0)
 	ttyname = buf;
 
@@ -1075,6 +1156,8 @@ sterm_sub_open(void *handler_data, struct gensio_iod **riod,
 	options |= GENSIO_OPEN_OPTION_WRITEABLE;
     if (!sdata->write_only)
 	options |= GENSIO_OPEN_OPTION_READABLE;
+    if (sdata->set_tty)
+	options |= GENSIO_OPEN_OPTION_SERIAL;
     err = o->open_dev(o, sdata->devname, options, &sdata->iod);
     if (err)
 	goto out;
@@ -1176,7 +1259,7 @@ sterm_sub_open(void *handler_data, struct gensio_iod **riod,
     sterm_unlock(sdata);
 
     if (sdata->set_tty)
-	sterm_modemstate(sdata, 255, NULL);
+	sterm_modemstate_mask(sdata, 255, NULL, NULL, NULL);
 
     *riod = sdata->iod;
 
@@ -1201,7 +1284,12 @@ static void
 sterm_free(void *handler_data)
 {
     struct sterm_data *sdata = handler_data;
+    struct modemstate_cb *c, *n;
 
+    for (c = sdata->modemstate_cbs; c; c = n) {
+	n = c->next;
+	sdata->o->free(sdata->o, c);
+    }
     if (sdata->sio)
 	sergensio_data_free(sdata->sio);
     serconf_clear_q(sdata);
@@ -1360,7 +1448,15 @@ sterm_control(void *handler_data, struct gensio_iod *iod,
     case GENSIO_CONTROL_SER_MODEMSTATE:
 	if (get)
 	    return GE_NOTSUP;
-	return sterm_modemstate(sdata, 0, data);
+	return sterm_modemstate_mask(sdata, 0, data, NULL, NULL);
+
+    case GENSIO_CONTROL_SER_LINESTATE:
+	if (!sdata->o->read_flags)
+	    return GE_NOTSUP;
+	if (get)
+	    return GE_NOTSUP;
+
+	return sterm_linestate_mask(sdata, 0, data, NULL, NULL);
 
     case GENSIO_CONTROL_SER_FLOWCONTROL_STATE:
 	if (get)
@@ -1396,11 +1492,25 @@ sterm_write(void *handler_data, struct gensio_iod *iod, gensiods *rcount,
 }
 
 static int
-sterm_do_read(struct gensio_iod *iod, void *data, gensiods count, gensiods *rcount,
-	      const char ***auxdata, void *cb_data)
+sterm_do_read(struct gensio_iod *iod, void *data, gensiods datalen,
+	      gensiods *rcount, const char ***auxdata, void *cb_data)
 {
     struct sterm_data *sdata = cb_data;
-    int rv = sdata->o->read(iod, data, count, rcount);
+    int rv;
+
+    if (sdata->o->read_flags) {
+	gensiods pcount = datalen / 2, count;
+
+	rv = sdata->o->read_flags(iod, data, data + pcount, pcount, &count);
+	if (!rv) {
+	    /* We put the flags into the second half of the data. */
+	    memmove(data + count, data + pcount, count);
+	    if (rcount)
+		*rcount = count * 2;
+	}
+    } else {
+	rv = sdata->o->read(iod, data, datalen, rcount);
+    }
 
     if (rv && sdata->is_pty && rv == GE_IOERR)
 	return GE_REMCLOSE; /* We don't seem to get EPIPE from ptys */
@@ -1415,6 +1525,82 @@ sterm_read_ready(void *handler_data, struct gensio_iod *iod)
     gensio_fd_ll_handle_incoming(sdata->ll, sterm_do_read, NULL, sdata);
 }
 
+static gensiods
+sterm_deliver_read(void *handler_data, void *cb_data, gensio_ll_cb cb, int err,
+		   void *data, gensiods datalen, const void *auxdata)
+{
+    struct sterm_data *sdata = handler_data;
+    gensiods total = 0;
+
+    if (!err && sdata->o->read_flags) {
+	gensiods i, start = 0, count, tosend, plen = datalen / 2;
+	unsigned char *buf = data, *flags = buf + plen;
+
+	for (i = 0; i < plen; i++) {
+	    if ((flags[i] & (GENSIO_IOD_READ_FLAGS_BREAK
+			     | GENSIO_IOD_READ_FLAGS_ERR)) != 0) {
+		struct gensio *io = sergensio_get_my_gensio(sdata->sio);
+		unsigned int linestate;
+		gensiods vlen = sizeof(linestate);
+
+		/* Send the data up to the point where the event is. */
+		tosend = i - start;
+		if (tosend > 0) {
+		    count = cb(cb_data, GENSIO_LL_CB_READ, 0,
+			       buf + start, tosend, NULL);
+		    if (count > tosend)
+			count = tosend;
+		    total += count;
+		    if (count < tosend)
+			/* Not all the data got written, give up. */
+			break;
+		}
+
+		/* Now send the event. */
+		linestate = 0;
+		if (flags[i] & GENSIO_IOD_READ_FLAGS_BREAK &&
+			sdata->linestate_mask & GENSIO_SER_LINESTATE_BREAK)
+		    linestate |= GENSIO_SER_LINESTATE_BREAK;
+		if (flags[i] & GENSIO_IOD_READ_FLAGS_ERR &&
+			sdata->linestate_mask & GENSIO_SER_LINESTATE_PARITY_ERR)
+		    linestate |= GENSIO_SER_LINESTATE_PARITY_ERR;
+
+		if (linestate)
+		    gensio_cb(io, GENSIO_EVENT_SER_LINESTATE, 0,
+			      (unsigned char *) &linestate, &vlen, NULL);
+
+		/* Note we don't report the bad data. */
+		start = i + 1;
+		total++;
+	    }
+	}
+
+	tosend = i - start;
+	if (tosend > 0) {
+	    count = cb(cb_data, GENSIO_LL_CB_READ, 0,
+		       buf + start, tosend, NULL);
+	    if (count > tosend)
+		count = tosend;
+	    total += count;
+	}
+	if (total < plen)
+	    /*
+	     * We didn't deliver all the data.  We have removed total
+	     * bytes from the beginning of the data and flags, so we
+	     * must shift the buffer over because the first "total * 2"
+	     * data bytes are going to be considered removed.
+	     */
+	    memmove(buf + total * 2, buf + total, plen - total);
+
+	total *= 2; /* Convert it back to the full amount. */
+    } else {
+	total = cb(cb_data, GENSIO_LL_CB_READ, err, data, datalen,
+		   auxdata);
+    }
+
+    return total;
+}
+
 static const struct gensio_fd_ll_ops sterm_fd_ll_ops = {
     .sub_open = sterm_sub_open,
     .check_close = sterm_check_close_drain,
@@ -1422,7 +1608,8 @@ static const struct gensio_fd_ll_ops sterm_fd_ll_ops = {
     .write = sterm_write,
     .read_ready = sterm_read_ready,
     .control = sterm_control,
-    .acontrol = sterm_acontrol
+    .acontrol = sterm_acontrol,
+    .deliver_read = sterm_deliver_read
 };
 
 static int
@@ -1785,6 +1972,13 @@ iodev_gensio_alloc(const void *gdata, const char * const args[],
     if (!sdata->lock)
 	goto out_nomem;
 
+    /*
+     * For the flags field to work, the buffer size must double sized,
+     * flags are held in the second half of the buffer.
+     */
+    if (o->read_flags)
+	max_read_size *= 2;
+
     sdata->ll = fd_gensio_ll_alloc(o, NULL, &sterm_fd_ll_ops, sdata,
 				   max_read_size, sdata->write_only,
 				   sdata->read_only);
