@@ -12,7 +12,7 @@ check_sctp()
 print("Test accept sctp")
 # FIXME - the raddr and laddr areq not tested here, it's hard to
 # know what it would be because of sctp multihoming.
-a = TestAccept(o, "sctp,ipv4,localhost,", "sctp,0", do_test, do_close = False)
+a = TestAccept(o, "sctp,ipv4,localhost,", "sctp,localhost,0", do_test, do_close = False)
 c = a.io1.control(0, gensio.GENSIO_CONTROL_GET,
                   gensio.GENSIO_CONTROL_STREAMS, None)
 if c != "instreams=1,ostreams=1":
