@@ -10,7 +10,7 @@ import gensio
 
 print("Test echo device")
 check_echo_dev()
-io = alloc_io(o, "serialdev," + ttyecho + ",38400")
+io = alloc_io(o, "serialdev," + ttyecho + ",38400,hangup-when-done")
 check_raddr(io, "echo device", ttyecho + ",38400N81 RTSHI DTRHI",
             ttyecho + ",38400N81,HANGUP_WHEN_DONE RTSHI DTRHI")
 test_dataxfer(io, io, "This is a test string!")
