@@ -20,7 +20,7 @@ def do_stream_test(io1, io2):
 
 print("Test sctp streams")
 TestAccept(o, "sctp(instreams=2,ostreams=3),localhost,",
-           "sctp(instreams=3,ostreams=2),0", do_stream_test,
+           "sctp(instreams=3,ostreams=2),localhost,0", do_stream_test,
            chunksize = 64)
 del o
 test_shutdown()
