@@ -13,8 +13,7 @@ check_pipe_dev()
 io1 = alloc_io(o, "serialdev," + ttypipe[0] + ",9600")
 io2 = alloc_io(o, "serialdev," + ttypipe[1] + ",9600")
 test_dataxfer(io1, io2, "This is a test string!")
-io_close(io1)
-io_close(io2)
+io_close((io1, io2))
 del io1
 del io2
 del o
