@@ -505,7 +505,7 @@ class HandleData:
             if (not self.expecting_modemstate_mask):
                 if (debug or self.debug):
                     print("Got unexpected modemstate mask for %s: %x" %
-                          (self.name, modemstate))
+                          (self.name, modemstate_mask))
                 self.enqueue("unexpected modemstate mask", modemstate_mask)
                 return
             if (modemstate_mask != self.expected_modemstate_mask):
@@ -567,6 +567,8 @@ class HandleData:
                 raise Exception("%s: Expecting linestate 0x%x, got 0x%x" %
                                 (self.name, self.expected_linestate,
                                  linestate))
+            if debug or self.debug:
+                print("Got linestate %x" % linestate)
             self.expecting_linestate = False
             self.wake("linestate", linestate)
         except Exception as e:
