@@ -147,7 +147,7 @@ list_sound_devs(struct gensio_os_funcs *o, const char *devtype)
 	err = gensio_read_s(lg, &len, buf, sizeof(buf) - 1, NULL);
 	if (!err) {
 	    buf[len] = '\0';
-	    puts(buf);
+	    fputs(buf, stdout);
 	}
     } while (!err);
 
@@ -247,10 +247,10 @@ main(int argc, char *argv[])
 	    list_devs = true;
 	} else if ((rv = cmparg(argc, argv, &arg, "-d", "--debug", NULL))) {
 	    gensio_set_log_mask(GENSIO_LOG_MASK_ALL);
-	} else if ((rv = cmparg(argc, argv, &arg, NULL, "--version", NULL))) {
+	} else if (cmparg(argc, argv, &arg, NULL, "--version", NULL)) {
 	    printf("Version %s\n", gensio_version_string);
 	    exit(0);
-	} else if ((rv = cmparg(argc, argv, &arg, "-h", "--help", NULL))) {
+	} else if (cmparg(argc, argv, &arg, "-h", "--help", NULL)) {
 	    help(0);
 	} else {
 	    fprintf(stderr, "Unknown argument: %s, us -h for help\n",
@@ -383,6 +383,7 @@ main(int argc, char *argv[])
     if (waiter)
 	gensio_os_funcs_free_waiter(o, waiter);
     gensio_os_proc_cleanup(proc_data);
+    gensio_cleanup_mem(o);
     gensio_os_funcs_free(o);
 
     return !!rv;
