@@ -27,6 +27,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <ctype.h>
 #include <assert.h>
 #ifdef HAVE_SETUTXENT
 #include <sys/time.h>
@@ -79,6 +80,7 @@ static unsigned int debug;
 static bool oneshot;
 static bool permit_root = false;
 static bool pw_login = false;
+static bool req_pw = false;
 static bool do_2fa = false;
 static bool ginteractive_login = true;
 static bool use_login = USE_LOGIN_PROGRAM;
@@ -190,6 +192,85 @@ glogger(void *cbdata, const char *format, ...)
     va_end(va);
 }
 
+static unsigned int gtlsshd_port = 852;
+
+#ifdef HAVE_MDNS
+#include <gensio/gensio_mdns.h>
+
+static bool enable_mdns = false;
+static const char *mdns_name;
+
+struct gensio_mdns *mdns;
+struct gensio_mdns_service *mdns_service;
+
+const char *mdns_txt[] = {
+    "provider=gtlsshd",
+    "gensiostack=mux.certauth,ssl,tcp",
+    NULL,
+};
+
+static void
+startup_mdns(struct gensio_os_funcs *o)
+{
+    int err = gensio_alloc_mdns(o, &mdns);
+    char *hostname = NULL;
+
+    if (!enable_mdns)
+	return;
+
+    if (!mdns_name) {
+	char *dot;
+	unsigned int i;
+
+	hostname = get_my_hostname(glogger, NULL);
+	if (!hostname)
+	    return;
+	dot = strchr(hostname, '.');
+	if (dot)
+	    *dot = '\0';
+	for (i = 0; hostname[i]; i++)
+	  hostname[i] = tolower((unsigned char) hostname[i]);
+	mdns_name = hostname;
+    }
+
+    if (err) {
+	log_event(LOG_ERR, "Unable to start MDNS: %s", gensio_err_to_str(err));
+	goto out_free;
+    }
+    err = gensio_mdns_add_service(mdns, -1,
+				  GENSIO_NETTYPE_UNSPEC,
+				  mdns_name, "_iostream._tcp",
+				  NULL, NULL,
+				  gtlsshd_port, mdns_txt,
+				  &mdns_service);
+    if (err) {
+	log_event(LOG_ERR, "Unable to add MDNS service: %s",
+		  gensio_err_to_str(err));
+	gensio_free_mdns(mdns, NULL, NULL);
+	mdns = NULL;
+    }
+ out_free:
+    if (hostname)
+	free(hostname);
+}
+
+static void
+cleanup_mdns(void)
+{
+    if (mdns_service) {
+	gensio_mdns_remove_service(mdns_service);
+	mdns_service = NULL;
+    }
+    if (mdns) {
+	gensio_free_mdns(mdns, NULL, NULL);
+	mdns = NULL;
+    }
+}
+#else
+static void startup_mdns(struct gensio_os_funcs *o) { }
+static void cleanup_mdns(void) { }
+#endif
+
 /* Default the program to this path. */
 #define STANDARD_PATH "/usr/local/bin:/usr/bin:/bin:/usr/sbin:/sbin"
 
@@ -242,7 +323,12 @@ struct auth_data {
     gensiods len_2fa;
 
     char *raddr; /* From GETSIO_CONTROL_RADDR */
+    char *laddr; /* From GETSIO_CONTROL_LADDR */
+
+    char *lhost; /* Host part, points into laddr, don't free */
+    char *lport;
     char *rhost; /* Host part, points into raddr, don't free */
+    char *rport;
     bool authed_by_cert;
 
     struct gtlssh_aux_data aux_data;
@@ -501,6 +587,8 @@ auth_free(struct auth_data *auth)
     }
     if (auth->raddr)
 	o->free(o, auth->raddr);
+    if (auth->laddr)
+	o->free(o, auth->laddr);
 #ifdef HAVE_LIBPAM
     if (auth->pam_session_open) {
 	auth->pam_err = pam_close_session(auth->pamh, PAM_SILENT);
@@ -602,6 +690,16 @@ close_con_info(struct per_con_info *pcinfo)
     auth_unlock(auth);
     if (pcinfo->io1_can_close) {
 	pcinfo->io1_can_close = false;
+#ifdef _WIN32
+	/*
+	 * Windows turns on win32-input-mode "\e[?9001h" and focus
+	 * events "\e[?1004h" when it starts a pty, but it doesn't
+	 * appear to disable them.  Send the disable to clean things
+	 * up.
+	 */
+	if (pcinfo->is_pty)
+	    gensio_write(pcinfo->io1, NULL, "\e[?9001l\e[?1004l", 16, NULL);
+#endif
 	err = gensio_close(pcinfo->io1, io_close, pcinfo);
 	if (err) {
 	    log_event(LOG_ERR, "Unable to close remote: %s",
@@ -1079,7 +1177,7 @@ setup_auth(struct auth_data *auth)
     pn = progname;
     if (pam_service)
 	pn = pam_service;
-    else if (pam_cert_auth_progname && auth->authed_by_cert)
+    else if (pam_cert_auth_progname && (auth->authed_by_cert && !req_pw))
 	pn = pam_cert_auth_progname;
     auth->pam_conv.appdata_ptr = auth;
     auth->pam_err = pam_start(pn, auth->username, &auth->pam_conv, &auth->pamh);
@@ -1117,7 +1215,8 @@ finish_auth(struct auth_data *auth)
 	return GE_AUTHREJECT;
     }
 
-    if (!gensio_is_authenticated(auth->rem_io) || pam_cert_auth_progname) {
+    if (!gensio_is_authenticated(auth->rem_io) || req_pw ||
+		pam_cert_auth_progname) {
 	int tries = 3;
 
 	if (!auth->interactive_login)
@@ -1858,10 +1957,10 @@ switch_to_user(struct auth_data *auth)
 
 	CloseHandle(auth->userh);
 	auth->userh = NULL;
+	err = GetLastError();
 	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL,
-		      GetLastError(), 0, errbuf, sizeof(errbuf), NULL);
+		      err, 0, errbuf, sizeof(errbuf), NULL);
 	log_event(LOG_ERR, "Could not set thread user: %s", errbuf);
-	return gensio_os_err_to_err(auth->ginfo->o, GetLastError());
     }
  out_win_err:
     if (err)
@@ -2086,7 +2185,7 @@ certauth_event(struct gensio *io, void *user_data, int event, int ierr,
 	     * don't log on with a password, and we don't want to have
 	     * to ask for one on certificate logons.
 	     */
-	    {
+	    if (!req_pw) {
 		char *pwfile = alloc_sprintf("%s%c.gtlssh%cpassword",
 					     auth->homedir, DIRSEP, DIRSEP);
 		size_t flen = 100;
@@ -2159,6 +2258,9 @@ new_rem_io(struct gensio *io, struct auth_data *auth)
     unsigned int env_len = 0;
     const char **penv2 = NULL;
     bool rv = false;
+#ifndef _WIN32
+    char *cmdbuf = NULL;
+#endif
 
     len = 0;
     err = gensio_control(io, 0, GENSIO_CONTROL_GET, GENSIO_CONTROL_SERVICE,
@@ -2184,10 +2286,25 @@ new_rem_io(struct gensio *io, struct auth_data *auth)
 	goto out_free;
     }
     if (strstartswith(service, "program:")) {
-	char *str = strchr(service, ':') + 1;
+	char *str = strchr(service, ':'), **svals;
+#ifdef _WIN32
+	unsigned int nsvals;
+#else
+	gensiods pos, alloclen;
+	char dummy[1];
+#endif
 
+	if (!str) {
+	    gensio_time timeout = {10, 0};
+
+	    write_str_to_gensio("invalid program string",
+				io, &timeout, true);
+	    goto out_free;
+	}
+
+	str++;
 	len -= str - service;
-	err = get_vals_from_service(o, &progv, NULL, str, len);
+	err = get_vals_from_service(o, &svals, NULL, str, len);
     out_bad_vals:
 	if (err) {
 	    gensio_time timeout = {10, 0};
@@ -2196,6 +2313,56 @@ new_rem_io(struct gensio *io, struct auth_data *auth)
 				io, &timeout, true);
 	    goto out_free;
 	}
+#ifdef _WIN32
+	for (nsvals = 0; svals[nsvals]; nsvals++)
+	    ;
+	progv = o->zalloc(o, (3 + nsvals) * sizeof(*progv));
+	if (!progv) {
+	    free(svals);
+	    log_event(LOG_ERR, "Could not allocate progv memory");
+	    goto out_free;
+	}
+	progv[0] = auth->ushell;
+	progv[1] = "/c";
+	for (nsvals = 0; svals[nsvals]; nsvals++) {
+	    /*
+	     * We need a way to keep an empty parameter from looking
+	     * like the \0\0, so we add a '\xff' to the beginning of
+	     * an empty parameter, or any parameter starting with
+	     * '\0xff'.  We drop that here.
+	     */
+	    if (svals[nsvals][0] == '\xff')
+		progv[nsvals + 2] = svals[nsvals] + 1;
+	    else
+		progv[nsvals + 2] = svals[nsvals];
+	}
+	progv[nsvals + 2] = NULL;
+#else
+	progv = o->zalloc(o, 4 * sizeof(*progv));
+	if (!progv) {
+	    free(svals);
+	    log_event(LOG_ERR, "Could not allocate progv memory");
+	    goto out_free;
+	}
+
+	pos = 0;
+	alloclen = gensio_argv_snprintf(dummy, 0, &pos, (const char **) svals);
+	cmdbuf = o->zalloc(o, alloclen + 1);
+	if (!cmdbuf) {
+	    free(svals);
+	    log_event(LOG_ERR, "Could not allocate progv memory");
+	    goto out_free;
+	}
+	pos = 0;
+	gensio_argv_snprintf(cmdbuf, alloclen + 1, &pos, (const char **) svals);
+
+	progv[0] = auth->ushell;
+	progv[1] = "-c";
+	progv[2] = cmdbuf;
+	progv[3] = NULL;
+#endif
+	free(svals);
+
 	/* Dummy out the program, we will set it later with a control. */
 	s = gensio_alloc_sprintf(o,
 				 "stdio(stderr-to-stdout,readbuf=16384),dummy");
@@ -2261,78 +2428,6 @@ new_rem_io(struct gensio *io, struct auth_data *auth)
 	goto out_free;
     }
 
-    if (login || progv) {
-	unsigned int i = 0, j;
-
-	for (i = 0; auth->env && auth->env[i]; i++)
-	    ;
-#ifndef _WIN32
-	i += 4;
-#endif
-	i += env_len;
-	if (i == 0)
-	    goto skip_env;
-
-	penv2 = o->zalloc(o, (i + 1) * sizeof(char *));
-	if (!penv2) {
-	    log_event(LOG_ERR, "Failure to reallocate env for %s",
-		      auth->username);
-	    goto out_free;
-	}
-
-	i = 0;
-#ifndef _WIN32
-	penv2[i] = gensio_alloc_sprintf(o, "HOME=%s", auth->homedir);
-	if (!penv2[i]) {
-	    log_event(LOG_ERR, "Failure to alloc HOME env space for %s",
-		      auth->username);
-	    goto out_free;
-	}
-	i++;
-	penv2[i] = gensio_alloc_sprintf(o, "USER=%s", auth->username);
-	if (!penv2[i]) {
-	    log_event(LOG_ERR, "Failure to alloc USER env space for %s",
-		      auth->username);
-	    goto out_free;
-	}
-	i++;
-	penv2[i] = gensio_alloc_sprintf(o, "LOGNAME=%s", auth->username);
-	if (!penv2[i]) {
-	    log_event(LOG_ERR, "Failure to alloc LOGNAME env space for %s",
-		      auth->username);
-	    goto out_free;
-	}
-	i++;
-	penv2[i] = gensio_alloc_sprintf(o, "PATH=%s", STANDARD_PATH);
-	if (!penv2[i]) {
-	    log_event(LOG_ERR, "Failure to alloc PATH env space for %s",
-		      auth->username);
-	    goto out_free;
-	}
-	i++;
-#endif
-	for (j = 0; auth->env[j]; i++, j++) {
-	    penv2[i] = gensio_strdup(o, auth->env[j]);
-	    if (!penv2[i]) {
-		log_event(LOG_ERR, "Failure to alloc env space for %s",
-			  auth->username);
-		goto out_free;
-	    }
-	}
-	if (env) {
-	    for (j = 0; j < env_len; i++, j++) {
-		penv2[i] = gensio_strdup(o, env[j]);
-		if (!penv2[i]) {
-		    log_event(LOG_ERR, "Failure to alloc env space for %s",
-			      auth->username);
-		    goto out_free;
-		}
-	    }
-	}
-	penv2[i] = NULL;
-    }
- skip_env:
-
     pcinfo = o->zalloc(o, sizeof(*pcinfo));
     if (!pcinfo) {
 	log_event(LOG_ERR, "Unable to allocate SSL pc info");
@@ -2416,6 +2511,104 @@ new_rem_io(struct gensio *io, struct auth_data *auth)
 
     if (progv || login) {
 	char *use_env = (char *) env;
+	unsigned int i = 0, j;
+
+	for (i = 0; auth->env && auth->env[i]; i++)
+	    ;
+#ifndef _WIN32
+	i += 7;
+#endif
+	i += env_len;
+	if (i == 0)
+	    goto skip_env;
+
+	penv2 = o->zalloc(o, (i + 1) * sizeof(char *));
+	if (!penv2) {
+	    log_event(LOG_ERR, "Failure to reallocate env for %s",
+		      auth->username);
+	    goto out_free;
+	}
+
+	i = 0;
+#ifndef _WIN32
+	penv2[i] = gensio_alloc_sprintf(o, "HOME=%s", auth->homedir);
+	if (!penv2[i]) {
+	    log_event(LOG_ERR, "Failure to alloc HOME env space for %s",
+		      auth->username);
+	    goto out_free;
+	}
+	i++;
+	penv2[i] = gensio_alloc_sprintf(o, "USER=%s", auth->username);
+	if (!penv2[i]) {
+	    log_event(LOG_ERR, "Failure to alloc USER env space for %s",
+		      auth->username);
+	    goto out_free;
+	}
+	i++;
+	penv2[i] = gensio_alloc_sprintf(o, "LOGNAME=%s", auth->username);
+	if (!penv2[i]) {
+	    log_event(LOG_ERR, "Failure to alloc LOGNAME env space for %s",
+		      auth->username);
+	    goto out_free;
+	}
+	i++;
+	penv2[i] = gensio_alloc_sprintf(o, "PATH=%s", STANDARD_PATH);
+	if (!penv2[i]) {
+	    log_event(LOG_ERR, "Failure to alloc PATH env space for %s",
+		      auth->username);
+	    goto out_free;
+	}
+	i++;
+	/*
+	 * At least macos requires SSH_xxx variables to be set to detect
+	 * that this is an ssh login and correctly load bashrc for non
+	 * login connections.
+	 */
+	if (auth->rhost && auth->lport && auth->rport) {
+	    penv2[i] = gensio_alloc_sprintf(o, "SSH_CLIENT=%s %s %s",
+					    auth->rhost, auth->rport,
+					    auth->lport);
+	    if (!penv2[i]) {
+		log_event(LOG_ERR,
+			  "Failure to alloc SSH_CLIENT env space for %s",
+			  auth->username);
+		goto out_free;
+	    }
+	    i++;
+	}
+	if (auth->rhost && auth->lhost && auth->lport && auth->rport) {
+	    penv2[i] = gensio_alloc_sprintf(o, "SSH_CONNECTION=%s %s %s %s",
+					    auth->rhost, auth->rport,
+					    auth->lhost, auth->lport);
+	    if (!penv2[i]) {
+		log_event(LOG_ERR,
+			  "Failure to alloc SSH_CONNECTION env space for %s",
+			  auth->username);
+		goto out_free;
+	    }
+	    i++;
+	}
+#endif
+	for (j = 0; auth->env[j]; i++, j++) {
+	    penv2[i] = gensio_strdup(o, auth->env[j]);
+	    if (!penv2[i]) {
+		log_event(LOG_ERR, "Failure to alloc env space for %s",
+			  auth->username);
+		goto out_free;
+	    }
+	}
+	if (env) {
+	    for (j = 0; j < env_len; i++, j++) {
+		penv2[i] = gensio_strdup(o, env[j]);
+		if (!penv2[i]) {
+		    log_event(LOG_ERR, "Failure to alloc env space for %s",
+			      auth->username);
+		    goto out_free;
+		}
+	    }
+	}
+	penv2[i] = NULL;
+    skip_env:
 
 	if (penv2)
 	    use_env = (char *) penv2;
@@ -2470,7 +2663,11 @@ new_rem_io(struct gensio *io, struct auth_data *auth)
     if (penv2)
 	gensio_argv_free(o, penv2);
     if (progv)
-	o->free(o, progv); /* Entries are from the service string. */
+	o->free(o, progv); /* Entries are from the service string or cndbuf. */
+#ifndef _WIN32
+    if (cmdbuf)
+	o->free(o, cmdbuf);
+#endif
     if (service)
 	o->free(o, service);
     if (s)
@@ -2563,7 +2760,7 @@ handle_new(struct gensio *net_io)
     int err;
     const char *ssl_args[] = { ginfo->key, ginfo->cert, "mode=server", NULL };
     const char *certauth_args[] = { "mode=server", "allow-authfail", NULL,
-				    NULL, NULL };
+				    NULL, NULL, NULL };
     struct gensio *ssl_io = NULL, *certauth_io = NULL, *top_io = NULL;
     struct auth_data *auth;
     gensiods len;
@@ -2617,6 +2814,9 @@ handle_new(struct gensio *net_io)
     if (pw_login)
 	certauth_args[i++] = "enable-password";
 
+    if (req_pw)
+	certauth_args[i++] = "require-password";
+
     if (do_2fa)
 	certauth_args[i++] = "enable-2fa";
 
@@ -2688,14 +2888,15 @@ handle_new(struct gensio *net_io)
     if (err)
 	goto out_err;
 
-    /* Set rhost.  If any of thils fails, we just go on. */
+    /* Set rhost, raddr, and laddr.  If any of thils fails, we just go on. */
     len = 0;
     err = gensio_control(net_io, 0, GENSIO_CONTROL_GET,
 			 GENSIO_CONTROL_RADDR, &dummy, &len);
     if (!err && len > 0) {
 	char *c2;
 
-	auth->raddr = o->zalloc(o, len + 1);
+	len++;
+	auth->raddr = o->zalloc(o, len);
 	if (!auth->raddr)
 	    goto skip_rhost;
 	err = gensio_control(net_io, 0, GENSIO_CONTROL_GET,
@@ -2708,8 +2909,10 @@ handle_new(struct gensio *net_io)
 
 	/* Pull the address out, it's between the first and last comma. */
 	c2 = strrchr(auth->raddr, ',');
-	if (c2)
+	if (c2) {
 	    *c2 = '\0';
+	    auth->rport = c2 + 1;
+	}
 	c2 = strchr(auth->raddr, ',');
 	if (c2)
 	    c2++;
@@ -2718,11 +2921,50 @@ handle_new(struct gensio *net_io)
 
 	auth->rhost = c2;
     }
+    len = 0;
+    err = gensio_control(net_io, 0, GENSIO_CONTROL_GET,
+			 GENSIO_CONTROL_LADDR, &dummy, &len);
+    if (!err && len > 0) {
+	char *c2;
+
+	len++;
+	auth->laddr = o->zalloc(o, len);
+	if (!auth->laddr)
+	    goto skip_rhost;
+	err = gensio_control(net_io, 0, GENSIO_CONTROL_GET,
+			     GENSIO_CONTROL_LADDR, auth->laddr, &len);
+	if (err) {
+	    o->free(o, auth->laddr);
+	    auth->laddr = NULL;
+	    goto skip_rhost;
+	}
+
+	/* Pull the address out, it's between the first and last comma. */
+	c2 = strrchr(auth->laddr, ',');
+	if (c2) {
+	    *c2 = '\0';
+	    auth->lport = c2 + 1;
+	}
+	c2 = strchr(auth->laddr, ',');
+	if (c2)
+	    c2++;
+	else
+	    c2 = auth->laddr;
+
+	auth->lhost = c2;
+    }
  skip_rhost:
 
     if (strstartswith(tmpservice, "login:"))
 	auth->interactive = true;
 
+    if (req_pw && !auth->passwd) {
+	log_event(LOG_ERR, "Password required, but none given for '%s'",
+		  auth->username);
+	err = GE_AUTHREJECT;
+	goto out_err;
+    }
+
     err = finish_auth(auth);
     if (err)
 	goto out_err;
@@ -2788,26 +3030,25 @@ setup_new_connection(struct gdata *ginfo, struct gensio *io)
 	return;
 
     case 0:
-	/*
-	 * The fork, let the parent have the accepter and double fork
-	 * so parent doesn't own us.  We have to tell the os handler,
-	 * too that we forked, or epoll() misbehaves.
-	 */
-	err = gensio_os_funcs_handle_fork(o);
-	if (err) {
-	    log_event(LOG_ERR, "Could not fork gensio handler: %s",
-		      gensio_err_to_str(err));
-	    exit(1);
-	    return;
-	}
-	pid_file = NULL; /* Make sure children don't delete this. */
-
 	setsid();
 	switch (fork()) {
 	case -1:
 	    log_event(LOG_ERR, "Could not fork twice: %s", strerror(errno));
 	    exit(1);
 	case 0:
+	    /*
+	     * The fork, let the parent have the accepter and double fork
+	     * so parent doesn't own us.  We have to tell the os handler,
+	     * too that we forked, or epoll() misbehaves.
+	     */
+	    err = gensio_os_funcs_handle_fork(o);
+	    if (err) {
+		log_event(LOG_ERR, "Could not fork gensio handler: %s",
+			  gensio_err_to_str(err));
+		exit(1);
+		return;
+	    }
+	    pid_file = NULL; /* Make sure children don't delete this. */
 	    break;
 	default:
 	    exit(0);
@@ -3062,6 +3303,7 @@ help(int err)
     printf("      Default is %s\n", keyfile);
     printf("  --permit-root - Allow root logins.\n");
     printf("  --allow-password - Allow password-based logins.\n");
+    printf("  --require-password - Use passwords even when certificates pass.\n");
     printf("  --oneshot - Do not fork new connections, do one and exit.\n");
     printf("  --nodaemon - Do not daemonize.\n");
     printf("  --nointeractive - Do not do interactive login queries.\n");
@@ -3083,6 +3325,12 @@ help(int err)
     printf("     %s login.\n", USE_LOGIN_PROGRAM ? "use" : "not use");
     printf("  --no-use-login - Don't use login program.\n");
     printf("  -P, --pidfile <file> - Create the given pidfile.\n");
+#endif
+#ifdef HAVE_MDNS
+    printf("  -m, --enable-mdns - Enable broadcasting MDNS information\n");
+    printf("     so MDNS users can find this.\n");
+    printf("  --mdns-name <name> - Set name used for MDNS.  The default\n");
+    printf("     is the first part of the hostname.\n");
 #endif
     printf("  --start-retries <count> - The number of retries for a name\n");
     printf("     lookup failure at startup.  The default is 30\n");
@@ -3131,7 +3379,6 @@ main(int argc, char *argv[])
     int arg, rv;
     struct gensio_os_funcs *o;
     struct gdata ginfo;
-    unsigned int port = 852;
     char *s;
     bool notcp = false, sctp = false;
     bool daemonize = true;
@@ -3153,7 +3400,7 @@ main(int argc, char *argv[])
 	    break;
 	}
 	if ((rv = cmparg_uint(argc, argv, &arg, "-p", "--port",
-			      &port)))
+			      &gtlsshd_port)))
 	    ;
 	else if ((rv = cmparg(argc, argv, &arg, "-c", "--certfile",
 			      &certfile)))
@@ -3172,6 +3419,8 @@ main(int argc, char *argv[])
 	    permit_root = true;
 	else if ((rv = cmparg(argc, argv, &arg, NULL, "--allow-password", NULL)))
 	    pw_login = true;
+	else if ((rv = cmparg(argc, argv, &arg, NULL, "--require-password", NULL)))
+	    req_pw = true;
 	else if ((rv = cmparg(argc, argv, &arg, NULL, "--do-2fa", NULL)))
 	    do_2fa = true;
 #ifdef HAVE_LIBPAM
@@ -3183,6 +3432,13 @@ main(int argc, char *argv[])
 	    ;
 	else if ((rv = cmparg(argc, argv, &arg, "-P", "--pidfile", &pid_file)))
 	    ;
+#endif
+#ifdef HAVE_MDNS
+	else if ((rv = cmparg(argc, argv, &arg, "-m", "--enable-mdns", NULL)))
+	    enable_mdns = true;
+	else if ((rv = cmparg(argc, argv, &arg, NULL, "--mdns-name",
+			      &mdns_name)))
+	    ;
 #endif
 	else if ((rv = cmparg(argc, argv, &arg, NULL, "--oneshot", NULL)))
 	    oneshot = true;
@@ -3288,7 +3544,8 @@ main(int argc, char *argv[])
 
  start_accept:
     if (!notcp && !tcp_acc) {
-	s = gensio_alloc_sprintf(o, "tcp(readbuf=20000),%s%d", iptype, port);
+	s = gensio_alloc_sprintf(o, "tcp(readbuf=20000),%s%d", iptype,
+				 gtlsshd_port);
 	if (!s) {
 	    log_event(LOG_ERR, "Could not allocate tcp descriptor\n");
 	    return 1;
@@ -3312,7 +3569,8 @@ main(int argc, char *argv[])
     }
 
     if (sctp && !sctp_acc) {
-	s = gensio_alloc_sprintf(o, "sctp(readbuf=20000),%s%d", iptype, port);
+	s = gensio_alloc_sprintf(o, "sctp(readbuf=20000),%s%d", iptype,
+				 gtlsshd_port);
 	if (!s) {
 	    log_event(LOG_ERR, "Could not allocate sctp descriptor\n");
 	    return 1;
@@ -3370,8 +3628,12 @@ main(int argc, char *argv[])
     if (!oneshot && daemonize)
 	do_daemonize(o);
 
+    startup_mdns(o);
+
     gensio_os_funcs_wait(o, ginfo.waiter, 1, NULL);
 
+    cleanup_mdns();
+
     /* FIXME - shutdown threads first. */
 
     if (tcp_acc) {
